/*******************************************************************************

                     GGGGG   QQQQQ  UU   UU  AAAAA  DDDDDD
                    GG   GG QQ   QQ UU   UU AA   AA  DD  DD
                    GG   GG QQ   QQ UU   UU AA   AA  DD  DD
                    GG      QQ   QQ UU   UU AAAAAAA  DD  DD
                    GG GGGG QQ   QQ UU   UU AA   AA  DD  DD
                    GG   GG QQ   QQ UU   UU AA   AA  DD  DD
                     GGGGG   QQQQQ   UUUUU  AA   AA DDDDDD
                    ------------QQ-------------------------
                    Gauss Quadrature Coefficients and Code
                  by Dudley J. Benton, dudley.benton@gmail.com

  GQUAD computes the coefficients (weights and abscissas) for Gauss
  Quadrature - the most efficient algorithm there is for numerical
  integration.  GQUAD generates the C and FORTRAN code for you - ready to
  use!  Just specify the order and redirect the output to a file as in:

       GQUAD 20 >INTEGRATE.FOR
       GQUAD 99 >integrate.c

  Gauss quadrature is quite superior to every other method.  It can be
  rigorously proven that Gauss quadrature is the most efficient algorithm
  of its kind.  GQUAD computes the weights and abscissas without resorting
  to any matrix operations.  It's all done with orthogonal polynomials.

--History of Revisions ---------------------------------------------------------

 V1.00 (05/18/86): original version written in assembler for real*8
 V2.00 (08/10/93): convert to C and real*10
 V2.01 (12/19/97): P2T distribution version
 V2.02 (03/11/99): fix bug in missing A[0]=0 for odd cases

*******************************************************************************/

char Version[]="2.02";

#define _CRT_SECURE_NO_DEPRECATE
#include <stdio.h>
#include <stddef.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <malloc.h>
#define _USE_MATH_DEFINES
#include <float.h>
#include <math.h>

typedef long double DBL;

#define tiny (100*LDBL_EPSILON)

void Abort(int code,char*message,...)
  {
  va_list arg_marker;
  fprintf(stderr,"\nFatal Error Code %li\n",code);
  va_start(arg_marker,message);
  vfprintf(stderr,message,arg_marker);
  fprintf(stderr,"\n");
  exit(1);
  }

void LegendrePolynomial(int n,DBL X,DBL*P3,DBL*Q3)
  {
  int i;
  DBL P1,P2,Q1,Q2;

  *P3=1;
  *Q3=0;
  if(n<1)
    return;

  P2=*P3;
  Q2=*Q3;
  *P3=X;
  *Q3=1;
  if(n<2)
    return;

  for(i=2;i<=n;i++)
    {
    P1=P2;
    Q1=Q2;
    P2=*P3;
    Q2=*Q3;
    *P3=2*X*P2-P1-(X*P2-P1)/i;
    *Q3=Q1+(2*i-1)*P2;
    }
  }

DBL RefineRoot(int n,DBL X1,DBL X2)          /* find a root between X1 and X2 */
  {
  int i;
  DBL dP,P,X,Xo;
  X=(X1+X2)/2;
  for(i=0;i<20;i++)
    {
    LegendrePolynomial(n,X,&P,&dP);
    if(fabsl(dP)<.5)
      break;
    Xo=X;
    X=max(X1,min(X2,X-P/dP));
    if(fabsl(P)<=tiny||fabs(X-Xo)<=tiny)
      return(X);
    }
  return(2);
  }

int InsertRoot(DBL*Roots,int nr,DBL Root)             /* insert root in array */
  {
  int i,j;
  for(i=0;i<nr;i++)
    if(fabs(Roots[i]-Root)<tiny)
      return(0);
  for(i=1;i<nr;i++)
    if(Roots[i-1]<Root&&Root<Roots[i])
      break;
  for(j=nr;j>i;j--)
    Roots[j]=Roots[j-1];
  Roots[i]=Root;
  return(i);
  }

int FindRoots(int n,DBL*Roots,DBL*Weights)
  {
  int i,j,k,nr;
  DBL dP,dX,P,Root,X,X1,X2;

  fprintf(stderr,"computing coefficients\n");

  if(n<2)
    return(0);

  nr=1;
  if(n%2)                                       /* find smallest root */
    Roots[0]=0;
  else
    {
    Roots[0]=RefineRoot(n,0,1.3/powl(n,.94));
    if(n<3)
      goto weights;
    }

  Roots[nr++]=RefineRoot(n,1-2.1/powl(n,1.9),1);/* find largest root */
  if(n<5)
    goto weights;

  fprintf(stderr,"\r%i roots found out of %i",nr,(n+1)/2);

  while(nr<(n+1)/2)      /* find all roots between each existing root */
    {
    for(i=1;i<nr;i++)
      {
      k=(n+1)/2-nr+1;
      dX=(Roots[i]-Roots[i-1])/k;
      X2=Roots[i-1];
      for(j=0;j<k;j++)
        {
        X1=X2;
        X2+=dX;
        Root=RefineRoot(n,X1,X2);
        if(Root>1.5)
          continue;
        if(!InsertRoot(Roots,nr,Root))
          continue;
        nr++;
        fprintf(stderr,"\r%i roots found out of %i",nr,(n+1)/2);
        break;
        }
      if(nr>=(n+1)/2)
        goto weights;
      }
    }

  weights:
  if(n>4)
    fprintf(stderr,"\n");
  for(i=0;i<(n+1)/2;i++)
    {
    X=Roots[i];
    LegendrePolynomial(n,X,&P,&dP);
    Weights[i]=2/(1-X*X)/dP/dP;
    }
  return((n+1)/2);
  }

void CreateCode(int n,DBL*Roots,DBL*Weights)
  {
  int i,w;

  fprintf(stderr,"generating C source code\n");

  w=(n+1)/2;

  printf("#include <float.h>\n");
  printf("#include <math.h>\n");
  printf("typedef long double DBL;\n\n");

  printf("DBL GQ%i(DBL F(DBL),DBL A,DBL B)      /",n);
  printf("* %i-pt Gauss qaudrature *",n);
  printf("/\n  {\n");

  if(n==2)
    printf("  static DBL X0=%21.19Lf;\n",Roots[0]);
  else if(n==3)
    {
    printf("  static DBL X1=%21.19Lf;\n",Roots[1]);
    printf("  static DBL W0=%21.19Lf,W1=%21.19Lf;\n",Weights[0],Weights[1]);
    }
  else
    {
    printf("  static DBL X[%i]={",w);
    if(n!=4)
      printf("        ");
    for(i=0;i<w;i++)
      {
      printf("%21.19Lf",Roots[i]);
      if(i!=w-1)
        {
        printf(",");
        if((i+2)%3==0)
          printf("\n    ");
        }
      else
        printf("};\n");
      }
    printf("  static DBL W[%i]={",w);
    if(n!=4)
      printf("        ");
    for(i=0;i<w;i++)
      {
      printf("%21.19Lf",Weights[i]);
      if(i!=w-1)
        {
        printf(",");
        if((i+2)%3==0)
          printf("\n    ");
        }
      else
        printf("};\n");
      }
    }

  if(n==2)
    {
    printf("  DBL C,D;\n");
    printf("  C=(A+B)/2;\n");
    printf("  D=(B-A)/2;\n");
    printf("  return((F(C-D*X0)+F(C+D*X0))*D);\n");
    }
  else if(n==3)
    {
    printf("  DBL C,D;\n");
    printf("  C=(A+B)/2;\n");
    printf("  D=(B-A)/2;\n");
    printf("  return((W0*F(C)+W1*(F(C-D*X1)+F(C+D*X1)))*D);\n");
    }
  else
    {
    printf("  int i;\n");
    printf("  DBL C,D,Q;\n");
    printf("  C=(A+B)/2;\n");
    printf("  D=(B-A)/2;\n");
    if(n%2)
      {
      printf("  Q=W[0]*F(C);\n");
      printf("  for(i=1;i<%i;i++)\n",w);
      printf("    Q+=W[i+1]*(F(C-D*X[i])+F(C+D*X[i]));\n");
      }
    else
      {
      printf("  Q=0;\n");
      printf("  for(i=0;i<%i;i++)\n",w);
      printf("    Q+=W[i]*(F(C-D*X[i])+F(C+D*X[i]));\n");
      }
    printf("  return(Q*D);\n");
    }
  printf("  }\n");
  }

void FortranCode(int n,DBL*Roots,DBL*Weights)
  {
  int i,w;

  fprintf(stderr,"generating FORTRAN source code\n");

  w=(n+1)/2;

  printf("      FUNCTION GQ%i(F,A,B)\n",n);
  printf("C %i-pt Gauss qaudrature\n",n);
  printf("      IMPLICIT INTEGER*4(I-M),REAL*8(A-H,O-Z)\n",n);

  if(n==2)
    printf("      DATA X0/%21.19Lf/\n",Roots[0]);
  else if(n==3)
    {
    printf("      DATA X1/%21.19Lf/\n",Roots[1]);
    printf("      DATA W0,W1/%21.19Lf,%21.19Lf/\n",Weights[0],Weights[1]);
    }
  else
    {
    printf("      DIMENSION X(%i),W(%i)\n",w,w);
    printf("      DATA X/");
    if(n!=4)
      printf("               ");
    for(i=0;i<w;i++)
      {
      printf("%21.19Lf",Roots[i]);
      if(i!=w-1)
        {
        printf(",");
        if((i+2)%3==0)
          printf("\n     &");
        }
      else
        printf("/\n");
      }
    printf("      DATA W/");
    if(n!=4)
      printf("               ");
    for(i=0;i<w;i++)
      {
      printf("%21.19Lf",Weights[i]);
      if(i!=w-1)
        {
        printf(",");
        if((i+2)%3==0)
          printf("\n     &");
        }
      else
        printf("/\n");
      }
    }

  printf("      C=(A+B)/2\n");
  printf("      D=(B-A)/2\n");
  if(n==2)
    printf("      GQ%i=(F(C-D*X0)+F(C+D*X0))*D\n",n);
  else if(n==3)
    printf("      GQ%i=(W0*F(C)+W1*(F(C-D*X1)+F(C+D*X1)))*D\n",n);
  else
    {
    if(n%2)
      {
      printf("      GQ%i=W(1)*F(C)\n",n);
      printf("      DO I=2,%i\n",w);
      printf("        GQ%i=GQ%i+W(I)*(F(C-D*X(I))+F(C+D*X(I)))\n",n,n);
      printf("      ENDDO\n");
      }
    else
      {
      printf("      GQ%i=0\n",n);
      printf("      DO I=1,%i\n",w);
      printf("        GQ%i=GQ%i+W(I)*(F(C-D*X(I))+F(C+D*X(I)))\n",n,n);
      printf("      ENDDO\n");
      }
    printf("      GQ%i=GQ%i*D\n",n,n);
    }
  printf("      RETURN\n");
  printf("      END\n");
  }

void ExpandRoots(int N,DBL*An,DBL*Wn)         /* This section of code expands */
  {                                          /* the compact/symmetric form of */
  int i,n,odd;                                     /* the GQ data statements. */
  DBL A;

  fprintf(stderr,"generating list of raw coefficients\n");

  odd=N%2;

  i=printf("\nDBL A%i[%i]={",N,N);
  while(i<25)
    printf(" ",i++);

  for(n=0;n<N;n++)
    {
    if(2*n==N-1)
      A=.5;
    else if(n<N/2)
      {
      i=N/2-n-1+odd;
      A=(1-An[i])/2;
      }
    else
      {
      i=n-N/2;
      A=(1+An[i])/2;
      }

    if(n%3==2)
      printf("\n  ");
    if(n==N-1)
      printf("%21.19Lf};\n",A);
    else
      printf("%21.19Lf,",A);
    }

  i=printf("\nDBL W%i[%i]={",N,N);
  while(i<25)
    printf(" ",i++);

  for(n=0;n<N;n++)
    {
    if(2*n==N-1)
      i=0;
    else if(n<N/2)
      i=N/2-n-1+odd;
    else
      i=n-N/2;

    if(n%3==2)
      printf("\n  ");
    if(n==N-1)
      printf("%21.19Lf};\n\n",Wn[i]/2);
    else
      printf("%21.19Lf,",Wn[i]/2);
    }
  }

void Usage(char*message)
  {
  fprintf(stderr,"\n"
    "This program computes the coefficients (weights and abscissas)\n"
    "for Gauss Quadrature - the most efficient algorithm there is for\n"
    "numerical integration.  GQUAD generates the C and FORTRAN code for\n"
    "you - ready to use!  Just specify the order and redirect the output\n"
    "to a file as in:\n"
    "\n"
    "        GQUAD 20 >INTEGRATE.FOR\n"
    "        GQUAD 99 >integrate.c\n"
    "\nError: %s\n",message);
  exit(1);
  }

int main(int argc,char**argv,char**envp)
  {
  int No;
  DBL*Roots,*Weights;

  fprintf(stderr,"GQUAD/V%s: Gauss Quadrature Integration Coefficients\n",Version);
  fprintf(stderr,"by Dudley J. Benton, Knoxville, Tennessee dudley.benton@gmail.com\n");

  if(argc==1)
    Usage("no parameters");

  if(argc!=2)
    Usage("wrong number of parameters");

  if(sscanf(argv[1],"%i",&No)!=1)
    Usage("scan error on parameter 1: order");

  if(No<2)
    Usage("order must be at least 2");

  if(No>999)
    Usage("order must not exceed 999");

  if((Roots=calloc(((No+1)/2),sizeof(DBL)))==NULL)
    Abort(__LINE__,"can't allocate memory");

  if((Weights=calloc(((No+1)/2),sizeof(DBL)))==NULL)
    Abort(__LINE__,"can't allocate memory");

  FindRoots(No,Roots,Weights);
  CreateCode(No,Roots,Weights);
  FortranCode(No,Roots,Weights);
  ExpandRoots(No,Roots,Weights);

  exit(0);
  }
